from typing import TYPE_CHECKING

from vital_ai_vitalsigns.vectordb.db.executors.typed_executor import TypedExecutor
from jina.serve.executors.decorators import requests, write

if TYPE_CHECKING:
    from docarray.index import InMemoryExactNNIndex
    from docarray import DocList


class InMemoryExactNNIndexer(TypedExecutor):

    def __init__(self, *args, **kwargs):
        from docarray.index import InMemoryExactNNIndex
        super().__init__(*args, **kwargs)
        self._index_file_path = f'{self.workspace}/index.bin' if self.handle_persistence else None
        self._indexer = InMemoryExactNNIndex[self._input_schema](index_file_path=self._index_file_path)

    def _index(self, docs, *args, **kwargs):
        self._indexer.index(docs)
        return docs

    @write
    @requests(on='/index')
    def index(self, docs, *args, **kwargs):
        self.logger.debug(f'Index {len(docs)}')
        return self._index(docs, *args, **kwargs)

    def _search(self, docs, parameters, *args, **kwargs):
        from docarray import DocList
        res = DocList[self._output_schema]()
        search_field = 'embedding'
        if parameters is not None and 'search_field' in parameters:
            search_field = parameters.pop('search_field')

        params = parameters or {}
        if '__results__' in params:
            params.pop('__results__')
        ret = self._indexer.find_batched(docs, search_field=search_field, **params)
        matched_documents = ret.documents
        matched_scores = ret.scores
        assert len(docs) == len(matched_documents) == len(matched_scores)

        for query, matches, scores in zip(docs, matched_documents, matched_scores):
            output_doc = self._output_schema(**query.dict())
            output_doc.matches = matches
            output_doc.scores = scores.tolist()
            res.append(output_doc)
        return res

    def _filter(self, docs, parameters, *args, **kwargs):
        from docarray import DocList
        res = DocList[self._output_schema]()
        search_field = 'embedding'
        class_uri = None

        if parameters is not None and 'search_field' in parameters:
            search_field = parameters.pop('search_field')

        if parameters is not None and 'ClassURI' in parameters:
            class_uri = parameters.pop('ClassURI')

        params = parameters or {}
        if '__results__' in params:
            params.pop('__results__')

        if class_uri is None:
            ret = self._indexer.find_batched(docs, search_field=search_field, **params)
        else:
            # filtered = self._indexer.filter(
            #    filter_query={'ClassURI': {'$eq': class_uri}},
            #    limit=self._indexer.num_docs())

            # this is not a batch case
            # TODO update for a list of class uris
            # check to see if limit is honored here
            # looks like it is in the find() function
            query = (self._indexer.build_query()
                     .filter(filter_query={'ClassURI': {'$eq': class_uri}})
                     .find(query=docs[0], search_field=search_field, limit=self._indexer.num_docs())
                     .build())

            # ret = self._indexer.find_batched(docs, search_field=search_field, **params)

            ret = self._indexer.execute_query(query)

            # self.logger.info(f'Results {ret}')

            # for d in ret:
            #    self.logger.info(f'{d}')

            matched_documents = ret.documents
            matched_scores = ret.scores

            output_doc = self._output_schema(**docs[0].dict())
            output_doc.matches = matched_documents
            output_doc.scores = matched_scores
            res.append(output_doc)
            return res

        # batch case
        matched_documents = ret.documents
        matched_scores = ret.scores

        # not true for filtered?
        # assert len(docs) == len(matched_documents) == len(matched_scores)

        for query, matches, scores in zip(docs, matched_documents, matched_scores):
            output_doc = self._output_schema(**query.dict())
            # self.logger.info(f'{query}')
            # self.logger.info(f'{matches}')
            # self.logger.info(f'{scores}')

            output_doc.matches = matches
            output_doc.scores = scores.tolist()
            res.append(output_doc)
        return res

    @requests(on='/search')
    def search(self, docs, *args, **kwargs):
        self.logger.debug(f'Search {len(docs)}')
        return self._search(docs, *args, **kwargs)

    def filter(self, docs, *args, **kwargs):
        self.logger.debug(f'Filter {len(docs)}')
        # self.logger.info(f'Filter {len(docs)}')
        return self._filter(docs, *args, **kwargs)

    def _delete(self, docs, *args, **kwargs):
        del self._indexer[[doc.id for doc in docs]]
        return docs

    @write
    @requests(on='/delete')
    def delete(self, docs, *args, **kwargs):
        self.logger.debug(f'Delete')
        return self._delete(docs, *args, **kwargs)

    @write
    @requests(on='/update')
    def update(self, docs, *args, **kwargs):
        # If not present, delete will raise IndexError and nothing will happen
        self._delete(docs)
        return self._index(docs)

    def num_docs(self, *args, **kwargs):
        return {'num_docs': self._indexer.num_docs()}

    def get_by_id(self, info_id, **kwargs):
        return self._indexer[info_id]

    def snapshot(self, snapshot_dir):
        snapshot_file = f'{snapshot_dir}/index.bin'
        self._indexer.persist(snapshot_file)

    def restore(self, snapshot_dir):
        from docarray.index import InMemoryExactNNIndex
        snapshot_file = f'{snapshot_dir}/index.bin'
        self._indexer = InMemoryExactNNIndex[self._input_schema](index_file_path=snapshot_file)

    def close(self):
        if self._index_file_path is not None:
            self._indexer.persist()
